package ca.dal.cs.schememm.common;

import java.util.Arrays;
import java.util.List;

import static org.testng.Assert.*;

public class CharRangesTest {

    @org.testng.annotations.Test
    public void testSubdivideLeft() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(0, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, 112));
        List<Pair<Integer, Integer>> subdivision =
                CharRanges.subdivide(charRanges);
        assertEquals(subdivision, Arrays.asList(new Pair<>(0, 26),
                                                new Pair<>(27, 32),
                                                new Pair<>(33, 34),
                                                new Pair<>(35, 36),
                                                new Pair<>(37, 39),
                                                new Pair<>(40, 42),
                                                new Pair<>(43, 112),
                                                new Pair<>(113,
                                                           (int) Character.MAX_VALUE)));
    }

    @org.testng.annotations.Test
    public void testSubdivideRight() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(2, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, (int) Character.MAX_VALUE));
        List<Pair<Integer, Integer>> subdivision =
                CharRanges.subdivide(charRanges);
        assertEquals(subdivision, Arrays.asList(new Pair<>(0, 1),
                                                new Pair<>(2, 26),
                                                new Pair<>(27, 32),
                                                new Pair<>(33, 34),
                                                new Pair<>(35, 36),
                                                new Pair<>(37, 39),
                                                new Pair<>(40, 42),
                                                new Pair<>(43, 0xffff)));
    }

    @org.testng.annotations.Test
    public void testSimplifyLeft() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(0, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, 112));
        List<Pair<Integer, Integer>> simpleRanges =
                CharRanges.simplify(charRanges);
        assertEquals(simpleRanges, Arrays.asList(new Pair<>(0, 34),
                                                 new Pair<>(37, 112)));
    }

    @org.testng.annotations.Test
    public void testSimplifyRight() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(2, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, 0xffff));
        List<Pair<Integer, Integer>> simpleRanges =
                CharRanges.simplify(charRanges);
        assertEquals(simpleRanges, Arrays.asList(new Pair<>(2, 34),
                                                 new Pair<>(37, 0xffff)));
    }

    @org.testng.annotations.Test
    public void testNegateLeft() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(0, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, 112));
        List<Pair<Integer, Integer>> negatedRanges =
                CharRanges.negate(charRanges);
        assertEquals(negatedRanges, Arrays.asList(new Pair<>(35, 36),
                                                  new Pair<>(113, 0xffff)));
    }

    @org.testng.annotations.Test
    public void testNegateRight() {
        List<Pair<Integer, Integer>> charRanges =
                Arrays.asList(new Pair<>(2, 32),
                              new Pair<>(27, 32),
                              new Pair<>(27, 34),
                              new Pair<>(37, 42),
                              new Pair<>(40, 0xffff));
        List<Pair<Integer, Integer>> negatedRanges =
                CharRanges.negate(charRanges);
        assertEquals(negatedRanges, Arrays.asList(new Pair<>(0, 1),
                                                  new Pair<>(35, 36)));
    }
}
