package ca.dal.cs.schememm.parser;

import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.scanner.Token;

import java.io.StringWriter;
import java.util.Iterator;
import java.util.Set;

class SyntaxError extends Exception {

    SyntaxError(Token token,
                Pair<Integer, Integer> pos, Set<Token.Name> expected) {
        this.token = token;
        this.pos = pos;
        this.expected = expected;
    }

    @Override
    public String getMessage() {
        Iterator<Token.Name> exp = expected.iterator();
        String               expectedStr;
        if (expected.size() == 1) {
            expectedStr = expectedStr(exp.next());
        }
        else if (expected.size() == 2) {
            expectedStr = String.format(
                    "%s or %s",
                    expectedStr(exp.next()),
                    expectedStr(exp.next()));
        }
        else {
            StringWriter w = new StringWriter();
            w.write(expectedStr(exp.next()));
            for (int i = 1; i < expected.size() - 1; ++i) {
                w.write(", ");
                w.write(expectedStr(exp.next()));
            }
            w.write(" or ");
            w.write(expectedStr(exp.next()));
            expectedStr = w.toString();
        }
        return String.format("Unexpected token '%s' ('%s') at position %d:%d" +
                             ".\n" +
                             "Expected %s.",
                             token.name(), token.string(),
                             pos.left, pos.right, expectedStr);
    }

    private String expectedStr(Token.Name name) {
        if (name == null) {
            return "end of input";
        }
        else {
            return "'" + name.toString() + "'";
        }
    }

    private Token                  token;
    private Pair<Integer, Integer> pos;
    private Set<Token.Name>        expected;
}
