package ca.dal.cs.schememm.parser.builder;

import ca.dal.cs.schememm.parser.ParserError;

import java.util.*;

/**
 * A class representing a grammar
 */
class Grammar {

    Grammar() {
        symbols = new HashMap<>();
        rules = new ArrayList<>();
    }

    /**
     * Add a rule to this grammar
     *
     * @param head       the head (left-hand side) of the rule
     * @param production the production (right-hand side) of the rule
     */
    void addRule(String head, String... production) throws ParserError {
        NonTerminal  headSym        = nonTerminal(head);
        List<Symbol> productionSyms = new ArrayList<>();
        for (String sym : production) {
            productionSyms.add(symbol(sym));
        }
        rules.add(new Rule(headSym, productionSyms));
    }

    /**
     * Return a non-terminal object with the given name.  All calls with the
     * same name return the same non-terminal object, so these objects can be
     * compared by identity.
     *
     * @param name the name of the non-terminal
     * @return the corresponding object
     */
    private NonTerminal nonTerminal(String name) throws ParserError {
        symbols.putIfAbsent(name, new NonTerminal(name));
        return symbols.get(name).asNonTerminal();
    }

    /**
     * Return a symbol object with the given name.  ALl calls with the same name
     * return the same symbol object, so these objects can be compared by
     * identity.
     *
     * @param name the name of the symbol
     * @return the corresponding object
     */
    private Symbol symbol(String name) throws ParserError {
        symbols.putIfAbsent(name, Symbol.symbol(name));
        return symbols.get(name);
    }

    /**
     * An iterator over the symbols in this grammar.
     */
    Collection<Symbol> symbols() {
        return symbols.values();
    }

    private Map<String, Symbol> symbols;
    List<Rule> rules;
}
