package ca.dal.cs.schememm.parser.builder;

import ca.dal.cs.schememm.parser.ParserError;
import ca.dal.cs.schememm.parser.parsetree.Node;
import ca.dal.cs.schememm.scanner.Token;

import java.util.*;

public class NonTerminal extends Symbol {

    /**
     * Construct a non-terminal in the grammar.  Checks that this is a valid
     * name of a non-terminal (the corresponding parse tree node class must
     * exist).
     *
     * @param name the name of the non-terminal
     */
    public NonTerminal(String name) throws ParserError {
        if (name.startsWith("V.")) {
            this.name = name.substring(2);
        }
        else {
            throw new ParserError(
                    String.format("%s is not a valid non-terminal", name));
        }
        followSet = new HashSet<>();
    }

    @Override
    public NonTerminal asNonTerminal() {
        return this;
    }

    /**
     * Return a const reference to this symbol's follow set
     *
     * @return the first set
     */
    Set<Token.Name> followSet() {
        return Collections.unmodifiableSet(followSet);
    }

    /**
     * Add more terminals to this symbol's follow set
     *
     * @param additions a list of character ranges to be added
     */
    boolean augmentFollowSet(Set<Token.Name> additions) {
        return followSet.addAll(additions);
    }

    @Override
    public Node node(ca.dal.cs.schememm.parser.parsetree.NonTerminal parent) {
        return new ca.dal.cs.schememm.parser.parsetree.NonTerminal(parent,
                                                                   name);
    }

    @Override
    public String name() {
        return name;
    }

    private final String          name;
    private       Set<Token.Name> followSet;
}
