package ca.dal.cs.schememm.parser.builder;

import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.parser.Parser;
import ca.dal.cs.schememm.parser.ParserError;
import ca.dal.cs.schememm.scanner.Token;

import java.io.StringWriter;
import java.util.*;

public class ParserBuilder {

    /**
     * Run the parser builder to obtain a parser
     *
     * @return the constructed parser
     */
    public static Parser run() throws ParserError {
        ParserBuilder builder = new ParserBuilder(grammar());
        builder.findFirstSets();
        builder.findFollowSets();
        builder.findPredictSets();
        return builder.parser();
    }

    /**
     * Static definition of the grammar for Scheme--
     *
     * @return the grammar
     */
    private static Grammar grammar() throws ParserError {
        Grammar grammar = new Grammar();
        grammar.addRule("V.Program");
        grammar.addRule("V.Program", "V.TopLevel", "V.Program");
        grammar.addRule("V.TopLevel",
                        "T.OPENRD", "V.DefOrFunCall", "T.CLOSERD");
        grammar.addRule("V.TopLevel",
                        "T.OPENSQ", "V.DefOrFunCall", "T.CLOSESQ");
        grammar.addRule("V.TopLevel",
                        "T.OPENCU", "V.DefOrFunCall", "T.CLOSECU");
        grammar.addRule("V.DefOrFunCall", "V.Definition1");
        grammar.addRule("V.DefOrFunCall", "V.FunCall1");
        grammar.addRule("V.Definition1", "T.DEFINE", "V.Definition2");
        grammar.addRule("V.Definition2", "T.IDENTIFIER", "V.Expression");
        grammar.addRule("V.Definition2", "V.ArgList", "V.Statements");
        grammar.addRule("V.ArgList",
                        "T.OPENRD", "T.IDENTIFIER", "V.ArgList1", "T.CLOSERD");
        grammar.addRule("V.ArgList",
                        "T.OPENSQ", "T.IDENTIFIER", "V.ArgList1", "T.CLOSESQ");
        grammar.addRule("V.ArgList",
                        "T.OPENCU", "T.IDENTIFIER", "V.ArgList1", "T.CLOSECU");
        grammar.addRule("V.ArgList1");
        grammar.addRule("V.ArgList1", "T.IDENTIFIER", "V.ArgList1");
        grammar.addRule("V.Statements", "V.SimpleExpression",
                        "V.MaybeStatements");
        grammar.addRule("V.Statements", "T.OPENRD", "V.DefOrCompoundRD");
        grammar.addRule("V.Statements", "T.OPENSQ", "V.DefOrCompoundSQ");
        grammar.addRule("V.Statements", "T.OPENCU", "V.DefOrCompoundCU");
        grammar.addRule("V.MaybeStatements");
        grammar.addRule("V.MaybeStatements", "V.Statements");
        grammar.addRule("V.DefOrCompoundRD", "V.Definition1", "T.CLOSERD",
                        "V.Statements");
        grammar.addRule("V.DefOrCompoundRD", "V.CompoundExpression1",
                        "T.CLOSERD", "V.MaybeStatements");
        grammar.addRule("V.DefOrCompoundSQ", "V.Definition1", "T.CLOSESQ",
                        "V.Statements");
        grammar.addRule("V.DefOrCompoundSQ", "V.CompoundExpression1",
                        "T.CLOSESQ", "V.MaybeStatements");
        grammar.addRule("V.DefOrCompoundCU", "V.Definition1", "T.CLOSECU",
                        "V.Statements");
        grammar.addRule("V.DefOrCompoundCU", "V.CompoundExpression1",
                        "T.CLOSECU", "V.MaybeStatements");
        grammar.addRule("V.FunCall1", "T.IDENTIFIER", "V.Expressions");
        grammar.addRule("V.Expressions");
        grammar.addRule("V.Expressions", "V.Expression", "V.Expressions");
        grammar.addRule("V.Expression", "V.SimpleExpression");
        grammar.addRule("V.Expression", "T.OPENRD", "V.CompoundExpression1",
                        "T.CLOSERD");
        grammar.addRule("V.Expression", "T.OPENSQ", "V.CompoundExpression1",
                        "T.CLOSESQ");
        grammar.addRule("V.Expression", "T.OPENCU", "V.CompoundExpression1",
                        "T.CLOSECU");
        grammar.addRule("V.SimpleExpression", "T.IDENTIFIER");
        grammar.addRule("V.SimpleExpression", "T.NUMBER");
        grammar.addRule("V.SimpleExpression", "T.CHAR");
        grammar.addRule("V.SimpleExpression", "T.BOOL");
        grammar.addRule("V.SimpleExpression", "T.STRING");
        grammar.addRule("V.SimpleExpression", "T.QUOTEMK", "V.Expression");
        grammar.addRule("V.CompoundExpression1", "V.FunCall1");
        grammar.addRule("V.CompoundExpression1", "T.LAMBDA",
                        "V.LambdaArgList", "V.Statements");
        grammar.addRule("V.CompoundExpression1", "T.QUOTE", "V.Expression");
        grammar.addRule("V.CompoundExpression1", "T.BEGIN", "V.Statements");
        grammar.addRule("V.CompoundExpression1", "T.LET", "V.LetBody");
        grammar.addRule("V.CompoundExpression1", "T.IF",
                        "V.Expression", "V.Expression", "V.Expression");
        grammar.addRule("V.CompoundExpression1", "T.COND", "V.CondBranches");
        grammar.addRule("V.LambdaArgList", "T.OPENRD", "V.ArgList1",
                        "T.CLOSERD");
        grammar.addRule("V.LambdaArgList", "T.OPENSQ", "V.ArgList1",
                        "T.CLOSESQ");
        grammar.addRule("V.LambdaArgList", "T.OPENCU", "V.ArgList1",
                        "T.CLOSECU");
        grammar.addRule("V.LetBody", "T.IDENTIFIER",
                        "V.VarDefs", "V.Statements");
        grammar.addRule("V.LetBody", "V.VarDefs", "V.Statements");
        grammar.addRule("V.VarDefs", "T.OPENRD", "V.VarDefs1", "T.CLOSERD");
        grammar.addRule("V.VarDefs", "T.OPENSQ", "V.VarDefs1", "T.CLOSESQ");
        grammar.addRule("V.VarDefs", "T.OPENCU", "V.VarDefs1", "T.CLOSECU");
        grammar.addRule("V.VarDefs1", "V.VarDef", "V.VarDefs2");
        grammar.addRule("V.VarDefs2");
        grammar.addRule("V.VarDefs2", "V.VarDef", "V.VarDefs2");
        grammar.addRule("V.VarDef", "T.OPENRD",
                        "T.IDENTIFIER", "V.Expression", "T.CLOSERD");
        grammar.addRule("V.VarDef", "T.OPENSQ",
                        "T.IDENTIFIER", "V.Expression", "T.CLOSESQ");
        grammar.addRule("V.VarDef", "T.OPENCU",
                        "T.IDENTIFIER", "V.Expression", "T.CLOSECU");
        grammar.addRule("V.CondBranches", "V.CondBranch", "V.CondBranches1");
        grammar.addRule("V.CondBranches1");
        grammar.addRule("V.CondBranches1", "V.CondBranch", "V.CondBranches1");
        grammar.addRule("V.CondBranch", "T.OPENRD",
                        "V.Expression", "V.Statements", "T.CLOSERD");
        grammar.addRule("V.CondBranch", "T.OPENSQ",
                        "V.Expression", "V.Statements", "T.CLOSESQ");
        grammar.addRule("V.CondBranch", "T.OPENCU",
                        "V.Expression", "V.Statements", "T.CLOSECU");
        return grammar;
    }

    /**
     * Construct a parser builder for the given grammar
     *
     * @param grammar the grammar for which to build a parser
     */
    private ParserBuilder(Grammar grammar) {
        this.grammar = grammar;
        startSymbol = grammar.rules.get(0).head.name();
        productions = new HashMap<>();
        expected = new HashMap<>();
    }

    /**
     * Find the first sets of all symbols in the grammar
     */
    private void findFirstSets() {
        boolean changed;
        do {
            changed = false;
            for (Rule rule : grammar.rules) {
                Set<Token.Name> firstSet = findFirstSet(rule.production);
                changed |= rule.head.augmentFirstSet(firstSet);
            }
        } while (changed);
    }

    /**
     * Find the first set of a sequence of symbols
     *
     * @param seq      the sequence of symbols
     * @return true if the entire sequence can vanish
     */
    private Set<Token.Name> findFirstSet(List<Symbol> seq) {
        Set<Token.Name> firstSet = new HashSet<>();
        for (Symbol sym : seq) {
            Set<Token.Name> symFirstSet = new HashSet<>(sym.firstSet());
            boolean canVanish = symFirstSet.contains(Token.Name.EOF);
            symFirstSet.remove(Token.Name.EOF);
            firstSet.addAll(symFirstSet);
            if (!canVanish) {
                return firstSet;
            }
        }
        firstSet.add(Token.Name.EOF);
        return firstSet;
    }

    /**
     * Find the follow sets of all non-terminals in the grammar
     */
    private void findFollowSets() {
        boolean changed;
        grammar.rules.get(0).head.augmentFollowSet(Collections.singleton(Token.Name.EOF));
        do {
            changed = false;
            for (Rule rule : grammar.rules) {
                Set<Token.Name> followSet =
                        new HashSet<>(rule.head.followSet());
                ListIterator<Symbol> symPtr =
                        rule.production.listIterator(rule.production.size());
                while (symPtr.hasPrevious()) {
                    Symbol sym = symPtr.previous();
                    NonTerminal var = sym.asNonTerminal();
                    if (var != null) {
                        changed |= var.augmentFollowSet(followSet);
                    }
                    Set<Token.Name> firstSet = new HashSet<>(sym.firstSet());
                    firstSet.remove(Token.Name.EOF);
                    if (sym.firstSet().contains(Token.Name.EOF)) {
                        followSet.addAll(firstSet);
                    }
                    else {
                        followSet = firstSet;
                    }
                }
            }
        } while (changed);
    }

    /**
     * Find the predict sets of all rules in the grammar
     */
    private void findPredictSets() {
        for (Rule rule : grammar.rules) {
            Set<Token.Name> firstSet = findFirstSet(rule.production);
            rule.predictSet = new HashSet<>(firstSet);
            rule.predictSet.remove(Token.Name.EOF);
            if (firstSet.contains(Token.Name.EOF)) {
                rule.predictSet.addAll(rule.head.followSet());
            }
        }
    }

    /**
     * The parser constructed by this parser builder
     */
    private Parser parser() throws ParserError {
        initProductions();
        storeProductions();
        buildExpectedSets();
        return new Parser(startSymbol, expected, productions);
    }

    /**
     * Makes sure that there's a (non-terminal, terminal) pair in the
     * productions lookup table for the given non-terminal and every terminal.
     */
    private void initProductions() {
        for (Symbol sym: grammar.symbols()) {
            NonTerminal var = sym.asNonTerminal();
            if (var != null) {
                for (Token.Name token : Token.Name.values()) {
                    productions.put(new Pair<>(var.name(), token), null);
                }
            }
        }
    }

    /**
     * Store the production to use for a given combination of non-terminal and
     * terminal, based on the terminals in the predict sets of the different
     * rules.
     */
    private void storeProductions() throws ParserError {
        for (Rule rule : grammar.rules) {
            for (Token.Name token : rule.predictSet) {
                Pair<String, Token.Name> key = new Pair<>(rule.head.name(),
                                                          token);
                if (productions.get(key) != null) {
                    ll1Conflict(rule.head.name(), token, rule);
                } else {
                    List<Symbol> production = new ArrayList<>(rule.production);
                    Collections.reverse(production);
                    productions.put(new Pair<>(rule.head.name(), token),
                                    production);
                }
            }
        }
    }

    /**
     * Record for every non-terminal at the beginning of the current sentential
     * form which terminals at the beginning of the input are acceptable.
     */
    private void buildExpectedSets() {
        for (Pair<String, Token.Name> pair : productions.keySet()) {
            expected.putIfAbsent(pair.left, new HashSet<>());
            if (productions.get(pair) != null) {
                expected.get(pair.left).add(pair.right);
            }
        }
    }

    /**
     * Report an LL(1) conflict, that is, a conflict between the predict sets of
     * two different productions for the same non-terminal.
     *
     * @param nonTerminal the non-terminal whose productions have a conflict
     * @param token       the token that occurs in multiple predict sets
     * @param rule        one of the rules that is involved in the conflict (the
     *                    other rule is already stored in the productions
     *                    table)
     */
    private void ll1Conflict(String nonTerminal, Token.Name token,
                             Rule rule) throws ParserError {
        StringWriter msg = new StringWriter();
        msg.write("The grammar is not LL(1)\n");
        msg.write(String.format("Conflicting token: %s\n", token.name()));
        msg.write("Conflicting productions:\n");
        msg.write(String.format("%s -->", nonTerminal));
        for (Symbol sym : rule.production) {
            msg.write(String.format(" %s", sym.name()));
        }
        msg.write(String.format("\n%s -->", nonTerminal));
        List<Symbol> production =
                new ArrayList<>(productions.get(new Pair<>(nonTerminal,
                                                           token)));
        ListIterator<Symbol> iter = production.listIterator(production.size());
        while (iter.hasPrevious()) {
            msg.write(String.format(" %s", iter.previous().name()));
        }
        throw new ParserError(msg.toString());
    }

    private Grammar                                     grammar;
    private String                                      startSymbol;
    private Map<String, Set<Token.Name>>                expected;
    private Map<Pair<String, Token.Name>, List<Symbol>> productions;
}
