package ca.dal.cs.schememm.parser.builder;

import ca.dal.cs.schememm.parser.ParserError;
import ca.dal.cs.schememm.parser.parsetree.Node;
import ca.dal.cs.schememm.scanner.Token;

import java.util.Collections;

public class Terminal extends Symbol {

    /**
     * Construct a terminal in the grammar.  Checks that this is a valid token
     * name of the scanner.
     *
     * @param name the token name
     */
    Terminal(String name) throws ParserError {
        if (name.startsWith("T.")) {
            try {
                token = Token.Name.valueOf(name.substring(2));
            }
            catch (IllegalArgumentException e) {
                throw new ParserError(
                        String.format("%s is not a valid token name", name));
            }
        }
        else {
            throw new ParserError(
                    String.format("%s is not a valid token name", name));
        }
        augmentFirstSet(Collections.singleton(token));
    }

    @Override
    public Node node(ca.dal.cs.schememm.parser.parsetree.NonTerminal parent) {
        return new ca.dal.cs.schememm.parser.parsetree.Terminal(parent, token);
    }

    @Override
    public String name() {
        return token.name();
    }

    private Token.Name token;
}
