package ca.dal.cs.schememm.parser.parsetree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * A parse tree node that represents a non-terminal, that is, an internal node.
 */
public class NonTerminal extends Node {

    public NonTerminal(NonTerminal parent, String name) {
        super(parent);
        this.name = name;
        children = new ArrayList<>();
    }

    /**
     * Add a child to this non-terminal node
     *
     * @param child the child to be added
     */
    void addChild(Node child) {
        children.add(child);
    }

    @Override
    public void format(String prefix) {
        System.out.printf("%s%s\n", prefix, name);
        if (prefix.length() > 0) {
            char junction = prefix.charAt(prefix.length() - 2);
            prefix = prefix.substring(0, prefix.length() - 2);
            if (junction == '├') {
                prefix += "│ ";
            } else {
                prefix += "  ";
            }
        }
        List<Node> children = new ArrayList<>(this.children);
        Collections.reverse(children);
        int numChildren = children.size();
        int i = 0;
        for (Node child : children) {
            if (++i == numChildren) {
                child.format(prefix + "└─");
            } else {
                child.format(prefix + "├─");
            }
        }
    }

    /**
     * The name of this non-terminal
     */
    public String name() {
        return name;
    }

    @Override
    public NonTerminal asNonTerminal() {
        return this;
    }

    private final String name;
    private List<Node> children;
}
