package ca.dal.cs.schememm.scanner;

import ca.dal.cs.schememm.common.CharStream;
import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.scanner.regex.RegexError;
import org.testng.annotations.Test;

import java.io.IOException;

import static org.testng.Assert.*;

public class ScannerTest {

    @Test
    public void success() throws IOException, RegexError, LexicalError {
        String input = "; This is a comment \n" +
                       "ident ; followed by an identifier\n" +
                       "     ; and another comment\n" +
                       "((][{)}[[} ; a few parentheses\n" +
                       "-143 +0 1443 0xbfa93 0x0 0x9 0b010110\n" +
                       "-0.314 +143.112 -27.0 .37 +.47 -.773 1e3 1e-3 1e+3\n" +
                       "1.003e-14 43.112e+13 .31e12 -.412e1 +.71e+0\n" +
                       "#t #f\n" +
                       "lambda (lambda define define() let cond if begin\n" +
                       "quote +ab a9 a-function\n" +
                       "' 'label '(a list) '14\n" +
                       "#\\a #\\9 #\\space #\\newline #\\tab #\\040\n" +
                       "\"A string \\\" \\t \\n \\040\"";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "ident");
        assertEquals(scanner.pos(), new Pair<>(2, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENRD);
        assertEquals(scanner.item().string(), "(");
        assertEquals(scanner.pos(), new Pair<>(4, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENRD);
        assertEquals(scanner.item().string(), "(");
        assertEquals(scanner.pos(), new Pair<>(4, 2));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSESQ);
        assertEquals(scanner.item().string(), "]");
        assertEquals(scanner.pos(), new Pair<>(4, 3));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENSQ);
        assertEquals(scanner.item().string(), "[");
        assertEquals(scanner.pos(), new Pair<>(4, 4));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENCU);
        assertEquals(scanner.item().string(), "{");
        assertEquals(scanner.pos(), new Pair<>(4, 5));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSERD);
        assertEquals(scanner.item().string(), ")");
        assertEquals(scanner.pos(), new Pair<>(4, 6));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSECU);
        assertEquals(scanner.item().string(), "}");
        assertEquals(scanner.pos(), new Pair<>(4, 7));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENSQ);
        assertEquals(scanner.item().string(), "[");
        assertEquals(scanner.pos(), new Pair<>(4, 8));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENSQ);
        assertEquals(scanner.item().string(), "[");
        assertEquals(scanner.pos(), new Pair<>(4, 9));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSECU);
        assertEquals(scanner.item().string(), "}");
        assertEquals(scanner.pos(), new Pair<>(4, 10));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "-143");
        assertEquals(scanner.pos(), new Pair<>(5, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "+0");
        assertEquals(scanner.pos(), new Pair<>(5, 6));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "1443");
        assertEquals(scanner.pos(), new Pair<>(5, 9));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "0xbfa93");
        assertEquals(scanner.pos(), new Pair<>(5, 14));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "0x0");
        assertEquals(scanner.pos(), new Pair<>(5, 22));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "0x9");
        assertEquals(scanner.pos(), new Pair<>(5, 26));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "0b010110");
        assertEquals(scanner.pos(), new Pair<>(5, 30));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "-0.314");
        assertEquals(scanner.pos(), new Pair<>(6, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "+143.112");
        assertEquals(scanner.pos(), new Pair<>(6, 8));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "-27.0");
        assertEquals(scanner.pos(), new Pair<>(6, 17));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), ".37");
        assertEquals(scanner.pos(), new Pair<>(6, 23));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "+.47");
        assertEquals(scanner.pos(), new Pair<>(6, 27));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "-.773");
        assertEquals(scanner.pos(), new Pair<>(6, 32));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "1e3");
        assertEquals(scanner.pos(), new Pair<>(6, 38));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "1e-3");
        assertEquals(scanner.pos(), new Pair<>(6, 42));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "1e+3");
        assertEquals(scanner.pos(), new Pair<>(6, 47));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "1.003e-14");
        assertEquals(scanner.pos(), new Pair<>(7, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "43.112e+13");
        assertEquals(scanner.pos(), new Pair<>(7, 11));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), ".31e12");
        assertEquals(scanner.pos(), new Pair<>(7, 22));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "-.412e1");
        assertEquals(scanner.pos(), new Pair<>(7, 29));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "+.71e+0");
        assertEquals(scanner.pos(), new Pair<>(7, 37));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.BOOL);
        assertEquals(scanner.item().string(), "#t");
        assertEquals(scanner.pos(), new Pair<>(8, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.BOOL);
        assertEquals(scanner.item().string(), "#f");
        assertEquals(scanner.pos(), new Pair<>(8, 4));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.LAMBDA);
        assertEquals(scanner.item().string(), "lambda");
        assertEquals(scanner.pos(), new Pair<>(9, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENRD);
        assertEquals(scanner.item().string(), "(");
        assertEquals(scanner.pos(), new Pair<>(9, 8));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.LAMBDA);
        assertEquals(scanner.item().string(), "lambda");
        assertEquals(scanner.pos(), new Pair<>(9, 9));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.DEFINE);
        assertEquals(scanner.item().string(), "define");
        assertEquals(scanner.pos(), new Pair<>(9, 16));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.DEFINE);
        assertEquals(scanner.item().string(), "define");
        assertEquals(scanner.pos(), new Pair<>(9, 23));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENRD);
        assertEquals(scanner.item().string(), "(");
        assertEquals(scanner.pos(), new Pair<>(9, 29));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSERD);
        assertEquals(scanner.item().string(), ")");
        assertEquals(scanner.pos(), new Pair<>(9, 30));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.LET);
        assertEquals(scanner.item().string(), "let");
        assertEquals(scanner.pos(), new Pair<>(9, 32));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.COND);
        assertEquals(scanner.item().string(), "cond");
        assertEquals(scanner.pos(), new Pair<>(9, 36));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IF);
        assertEquals(scanner.item().string(), "if");
        assertEquals(scanner.pos(), new Pair<>(9, 41));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.BEGIN);
        assertEquals(scanner.item().string(), "begin");
        assertEquals(scanner.pos(), new Pair<>(9, 44));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.QUOTE);
        assertEquals(scanner.item().string(), "quote");
        assertEquals(scanner.pos(), new Pair<>(10, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "+ab");
        assertEquals(scanner.pos(), new Pair<>(10, 7));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "a9");
        assertEquals(scanner.pos(), new Pair<>(10, 11));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "a-function");
        assertEquals(scanner.pos(), new Pair<>(10, 14));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.QUOTEMK);
        assertEquals(scanner.item().string(), "'");
        assertEquals(scanner.pos(), new Pair<>(11, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.QUOTEMK);
        assertEquals(scanner.item().string(), "'");
        assertEquals(scanner.pos(), new Pair<>(11, 3));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "label");
        assertEquals(scanner.pos(), new Pair<>(11, 4));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.QUOTEMK);
        assertEquals(scanner.item().string(), "'");
        assertEquals(scanner.pos(), new Pair<>(11, 10));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.OPENRD);
        assertEquals(scanner.item().string(), "(");
        assertEquals(scanner.pos(), new Pair<>(11, 11));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "a");
        assertEquals(scanner.pos(), new Pair<>(11, 12));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.IDENTIFIER);
        assertEquals(scanner.item().string(), "list");
        assertEquals(scanner.pos(), new Pair<>(11, 14));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CLOSERD);
        assertEquals(scanner.item().string(), ")");
        assertEquals(scanner.pos(), new Pair<>(11, 18));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.QUOTEMK);
        assertEquals(scanner.item().string(), "'");
        assertEquals(scanner.pos(), new Pair<>(11, 20));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.NUMBER);
        assertEquals(scanner.item().string(), "14");
        assertEquals(scanner.pos(), new Pair<>(11, 21));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\a");
        assertEquals(scanner.pos(), new Pair<>(12, 1));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\9");
        assertEquals(scanner.pos(), new Pair<>(12, 5));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\space");
        assertEquals(scanner.pos(), new Pair<>(12, 9));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\newline");
        assertEquals(scanner.pos(), new Pair<>(12, 17));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\tab");
        assertEquals(scanner.pos(), new Pair<>(12, 27));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.CHAR);
        assertEquals(scanner.item().string(), "#\\040");
        assertEquals(scanner.pos(), new Pair<>(12, 33));
        scanner.step();
        assertEquals(scanner.item().name, Token.Name.STRING);
        assertEquals(scanner.item().string(),
                     "\"A string \\\" \\t \\n \\040\"");
        assertEquals(scanner.pos(), new Pair<>(13, 1));
        scanner.step();
        assertNull(scanner.item());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void identifierFail1() throws IOException, RegexError, LexicalError {
        String     input   = "0ab";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void identifierFail2() throws IOException, RegexError, LexicalError {
        String     input   = "ab'cd";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void boolFail1() throws IOException, RegexError, LexicalError {
        String     input   = "#true";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void boolFail2() throws IOException, RegexError, LexicalError {
        String     input   = "#x";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void charFail1() throws IOException, RegexError, LexicalError {
        String     input   = "#\\aa";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void charFail2() throws IOException, RegexError, LexicalError {
        String     input   = "#\\tabular";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void charFail3() throws IOException, RegexError, LexicalError {
        String     input   = "#\\newlne";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void charFail4() throws IOException, RegexError, LexicalError {
        String     input   = "#\\3770";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void charFail5() throws IOException, RegexError, LexicalError {
        String     input   = "#\\477";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void stringFail1() throws IOException, RegexError, LexicalError {
        String     input   = "\"\"\"";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void stringFail2() throws IOException, RegexError, LexicalError {
        String     input   = "\"\\\"";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }

    @Test(expectedExceptions = {LexicalError.class})
    public void stringFail3() throws IOException, RegexError, LexicalError {
        String     input   = "\"\\a\"";
        CharStream chars   = CharStream.fromString(input);
        Scanner    scanner = new Scanner(chars);
        System.out.println(scanner.item().name());
        System.out.println(scanner.item().string());
    }
}
