package ca.dal.cs.schememm.scanner.fa;

import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.scanner.Token;
import ca.dal.cs.schememm.scanner.regex.RegexError;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.*;

import static org.testng.Assert.*;

public class DFATest {

    @Test
    public void testFromNFA() throws IOException, RegexError {
        NFA nfa1 = NFA.fromRegex("a?b+(c|d){1,2}", Token.Name.IDENTIFIER);
        NFA nfa2 = NFA.fromRegex("[a-z0-9]*[^a-z]", Token.Name.NUMBER);
        NFA nfa  = NFA.parallelJoin(Arrays.asList(nfa1, nfa2));
        DFA dfa  = DFA.fromNFA(nfa);

        State[] states     = new State[13];
        int numNeighbours;
        State[] neighbours = new State[9];

        states[0] = dfa.start;
        assertEquals(states[0].neighbours.size(), 7);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[0].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 122))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 7);
        assertSame(neighbours[0], neighbours[2]);
        states[1] = neighbours[0];
        states[3] = neighbours[1];
        states[6] = neighbours[3];
        states[7] = neighbours[4];
        states[4] = neighbours[5];
        states[5] = neighbours[6];

        assertEquals(states[1].neighbours.size(), 1);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[1].neighbours) {
            if (nbr.left.equals(new Pair<>(0, (int) Character.MAX_VALUE))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 1);
        states[2] = neighbours[0];

        assertEquals(states[2].neighbours.size(), 1);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[2].neighbours) {
            if (nbr.left.equals(new Pair<>(0, (int) Character.MAX_VALUE))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 1);
        assertSame(neighbours[0], states[2]);

        assertEquals(states[3].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[3].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[5]);

        assertEquals(states[4].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[4].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[5]);

        assertEquals(states[5].neighbours.size(), 1);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[5].neighbours) {
            if (nbr.left.equals(new Pair<>(0, (int) Character.MAX_VALUE))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 1);
        assertSame(neighbours[0], states[2]);

        assertEquals(states[6].neighbours.size(), 7);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[6].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 122))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 7);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[7]);
        assertSame(neighbours[5], states[4]);
        assertSame(neighbours[6], states[5]);

        assertEquals(states[7].neighbours.size(), 9);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[7].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 99))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(100, 100))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[7] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[8] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 9);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        states[8] = neighbours[4];
        states[9] = neighbours[5];
        states[12] = neighbours[6];
        assertSame(neighbours[7], states[4]);
        assertSame(neighbours[8], states[5]);

        assertEquals(states[8].neighbours.size(), 9);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[8].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 99))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(100, 100))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[7] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[8] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 9);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[8]);
        assertSame(neighbours[5], states[9]);
        assertSame(neighbours[6], states[12]);
        assertSame(neighbours[7], states[4]);
        assertSame(neighbours[8], states[5]);

        assertEquals(states[9].neighbours.size(), 8);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[9].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 98))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 99))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(100, 100))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[7] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 8);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        states[10] = neighbours[4];
        states[11] = neighbours[5];
        assertSame(neighbours[6], states[4]);
        assertSame(neighbours[7], states[5]);

        assertEquals(states[10].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[10].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[5]);

        assertEquals(states[11].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[11].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[5]);

        assertEquals(states[12].neighbours.size(), 8);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[12].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 98))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 99))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(100, 100))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[7] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 8);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[10]);
        assertSame(neighbours[5], states[11]);
        assertSame(neighbours[6], states[4]);
        assertSame(neighbours[7], states[5]);
    }

    @Test
    public void testMinimize() throws IOException, RegexError {
        NFA nfa1 = NFA.fromRegex("a?b+(c|d){1,2}", Token.Name.IDENTIFIER);
        NFA nfa2 = NFA.fromRegex("[a-z0-9]*[^a-z]", Token.Name.NUMBER);
        NFA nfa  = NFA.parallelJoin(Arrays.asList(nfa1, nfa2));
        DFA dfa  = DFA.fromNFA(nfa);
        DFA minDFA = DFA.minimize(dfa);

        State[] states     = new State[13];
        int numNeighbours;
        State[] neighbours = new State[9];

        states[0] = minDFA.start;
        assertEquals(states[0].neighbours.size(), 7);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[0].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 122))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 7);
        assertSame(neighbours[0], neighbours[2]);
        assertSame(neighbours[0], neighbours[6]);
        states[1] = neighbours[0];
        states[3] = neighbours[1];
        states[6] = neighbours[3];
        states[7] = neighbours[4];
        states[4] = neighbours[5];

        assertEquals(states[1].neighbours.size(), 1);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[1].neighbours) {
            if (nbr.left.equals(new Pair<>(0, (int) Character.MAX_VALUE))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 1);
        states[2] = neighbours[0];

        assertEquals(states[2].neighbours.size(), 1);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[2].neighbours) {
            if (nbr.left.equals(new Pair<>(0, (int) Character.MAX_VALUE))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 1);
        assertSame(neighbours[0], states[2]);

        assertEquals(states[3].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[3].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[1]);

        assertEquals(states[4].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[4].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[1]);

        assertEquals(states[6].neighbours.size(), 7);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[6].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 122))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 7);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[7]);
        assertSame(neighbours[5], states[4]);
        assertSame(neighbours[6], states[1]);

        assertEquals(states[7].neighbours.size(), 8);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[7].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 97))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(98, 98))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 100))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[7] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 8);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[7]);
        states[9] = neighbours[5];
        assertSame(neighbours[6], states[4]);
        assertSame(neighbours[7], states[1]);

        assertEquals(states[9].neighbours.size(), 7);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[9].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 98))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(99, 100))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(101, 122))) {
                neighbours[5] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[6] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 7);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        states[10] = neighbours[4];
        assertSame(neighbours[5], states[4]);
        assertSame(neighbours[6], states[1]);

        assertEquals(states[10].neighbours.size(), 5);
        numNeighbours = 0;
        for (Pair<Pair<Integer, Integer>, State> nbr : states[10].neighbours) {
            if (nbr.left.equals(new Pair<>(0, 47))) {
                neighbours[0] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(48, 57))) {
                neighbours[1] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(58, 96))) {
                neighbours[2] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(97, 122))) {
                neighbours[3] = nbr.right;
                ++numNeighbours;
            }
            if (nbr.left.equals(new Pair<>(123, (int) Character.MAX_VALUE))) {
                neighbours[4] = nbr.right;
                ++numNeighbours;
            }
        }
        assertEquals(numNeighbours, 5);
        assertSame(neighbours[0], states[1]);
        assertSame(neighbours[1], states[3]);
        assertSame(neighbours[2], states[1]);
        assertSame(neighbours[3], states[4]);
        assertSame(neighbours[4], states[1]);
    }

    @Test
    public void testRun() {
    }
}
