package ca.dal.cs.schememm.scanner.regex;

import ca.dal.cs.schememm.common.CharRanges;
import ca.dal.cs.schememm.common.Pair;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static org.testng.Assert.*;

public class RegexParserTest {

    @Test
    public void testIdentifier() throws IOException, RegexError {
        RegexParser parser = new RegexParser("[^][(){};0-9'\"#][^][(){};" +
                                             "'\"#]*");
        Regex regex = parser.run();
        List<Pair<Integer, Integer>> firstLetterRange =
                CharRanges.negate(
                        Arrays.asList(
                                new Pair<>((int) ']', (int) ']'),
                                new Pair<>((int) '[', (int) '['),
                                new Pair<>((int) '(', (int) '('),
                                new Pair<>((int) ')', (int) ')'),
                                new Pair<>((int) '{', (int) '{'),
                                new Pair<>((int) '}', (int) '}'),
                                new Pair<>((int) ';', (int) ';'),
                                new Pair<>((int) '0', (int) '9'),
                                new Pair<>((int) '\'', (int) '\''),
                                new Pair<>((int) '"', (int) '"'),
                                new Pair<>((int) '#', (int) '#')));
        List<Pair<Integer, Integer>> otherLetterRange =
                CharRanges.negate(
                        Arrays.asList(
                                new Pair<>((int) ']', (int) ']'),
                                new Pair<>((int) '[', (int) '['),
                                new Pair<>((int) '(', (int) '('),
                                new Pair<>((int) ')', (int) ')'),
                                new Pair<>((int) '{', (int) '{'),
                                new Pair<>((int) '}', (int) '}'),
                                new Pair<>((int) ';', (int) ';'),
                                new Pair<>((int) '\'', (int) '\''),
                                new Pair<>((int) '"', (int) '"'),
                                new Pair<>((int) '#', (int) '#')));
        Regex firstLetter = new AlternativeRegex(
                firstLetterRange.stream()
                                .map(rng -> new CharRangeRegex(rng.left,
                                                               rng.right))
                                .collect(Collectors.toList()));
        Regex otherLetter = new AlternativeRegex(
                otherLetterRange.stream()
                                .map(rng -> new CharRangeRegex(rng.left,
                                                               rng.right))
                                .collect(Collectors.toList()));
        Regex otherLetters = new RepeatRegex(otherLetter, 0, -1);
        Regex expected = new SequenceRegex(
                Arrays.asList(firstLetter, otherLetters));
        assertEquals(regex, expected);
    }

    @Test
    public void testInteger() throws IOException, RegexError {
        RegexParser parser =
                new RegexParser("[-+]?[0-9]+|0x[0-9a-fA-F]+|0b[01]+");
        Regex regex = parser.run();
        Regex sign = new RepeatRegex(
                new AlternativeRegex(
                        Arrays.asList(
                                new CharRangeRegex((int) '-', (int) '-'),
                                new CharRangeRegex((int) '+', (int) '+'))),
                0, 1);
        Regex digits = new RepeatRegex(
                new CharRangeRegex((int) '0', (int) '9'), 1, -1);
        Regex decimal = new SequenceRegex(Arrays.asList(sign, digits));
        Regex hexDigits = new RepeatRegex(
                new AlternativeRegex(
                        Arrays.asList(
                                new CharRangeRegex((int) '0', (int) '9'),
                                new CharRangeRegex((int) 'a', (int) 'f'),
                                new CharRangeRegex((int) 'A', (int) 'F'))),
                1, -1);
        Regex hex = new SequenceRegex(
                Arrays.asList(
                        new CharRangeRegex((int) '0', (int) '0'),
                        new CharRangeRegex((int) 'x', (int) 'x'),
                        hexDigits));
        Regex binDigits = new RepeatRegex(
                new AlternativeRegex(
                        Arrays.asList(
                                new CharRangeRegex((int) '0', (int) '0'),
                                new CharRangeRegex((int) '1', (int) '1'))),
                1, -1);
        Regex bin = new SequenceRegex(
                Arrays.asList(
                        new CharRangeRegex((int) '0', (int) '0'),
                        new CharRangeRegex((int) 'b', (int) 'b'),
                        binDigits));
        Regex expected = new AlternativeRegex(
                Arrays.asList(decimal, hex, bin));
        assertEquals(regex, expected);
    }

    @Test
    public void testFloat() throws IOException, RegexError {
        RegexParser parser =
                new RegexParser("[-+]?[0-9]*\\.[0-9]+([eE][-+]?[0-9]+)?|" +
                                "[-+]?[0-9]+[eE][-+]?[0-9]+");
        Regex regex = parser.run();
        Regex sign = new RepeatRegex(
                new AlternativeRegex(
                        Arrays.asList(
                                new CharRangeRegex((int) '-', (int) '-'),
                                new CharRangeRegex((int) '+', (int) '+'))),
                0, 1);
        Regex digit           = new CharRangeRegex((int) '0', (int) '9');
        Regex someDigits      = new RepeatRegex(digit, 1, -1);
        Regex maybeSomeDigits = new RepeatRegex(digit, 0, -1);
        Regex exponentE = new AlternativeRegex(
                Arrays.asList(
                        new CharRangeRegex((int) 'e', (int) 'e'),
                        new CharRangeRegex((int) 'E', (int) 'E')));
        Regex exponent = new SequenceRegex(
                Arrays.asList(exponentE, sign, someDigits));
        Regex period = new CharRangeRegex((int) '.', (int) '.');
        Regex option1 = new SequenceRegex(
                Arrays.asList(
                        sign, maybeSomeDigits, period, someDigits,
                        new RepeatRegex(exponent, 0, 1)));
        Regex option2 = new SequenceRegex(
                Arrays.asList(
                        sign, someDigits, exponentE, sign, someDigits));
        Regex expected = new AlternativeRegex(
                Arrays.asList(option1, option2));
        assertEquals(regex, expected);
    }

    @Test
    public void testBoolean() throws IOException, RegexError {
        RegexParser parser = new RegexParser("#[tf]");
        Regex       regex  = parser.run();
        Regex expected = new SequenceRegex(
                Arrays.asList(
                        new CharRangeRegex((int) '#', (int) '#'),
                        new AlternativeRegex(
                                Arrays.asList(
                                        new CharRangeRegex((int) 't',
                                                           (int) 't'),
                                        new CharRangeRegex((int) 'f',
                                                           (int) 'f')))));
        assertEquals(regex, expected);
    }

    @Test
    public void testCharacter() throws IOException, RegexError {
        RegexParser parser = new RegexParser(
                "#\\\\(.|newline|space|tab|[0-3][0-7]{2})");
        Regex regex     = parser.run();
        Regex n         = new CharRangeRegex((int) 'n', (int) 'n');
        Regex e         = new CharRangeRegex((int) 'e', (int) 'e');
        Regex w         = new CharRangeRegex((int) 'w', (int) 'w');
        Regex l         = new CharRangeRegex((int) 'l', (int) 'l');
        Regex i         = new CharRangeRegex((int) 'i', (int) 'i');
        Regex s         = new CharRangeRegex((int) 's', (int) 's');
        Regex p         = new CharRangeRegex((int) 'p', (int) 'p');
        Regex a         = new CharRangeRegex((int) 'a', (int) 'a');
        Regex c         = new CharRangeRegex((int) 'c', (int) 'c');
        Regex t         = new CharRangeRegex((int) 't', (int) 't');
        Regex b         = new CharRangeRegex((int) 'b', (int) 'b');
        Regex hash      = new CharRangeRegex((int) '#', (int) '#');
        Regex backslash = new CharRangeRegex((int) '\\', (int) '\\');
        Regex octFirst  = new CharRangeRegex((int) '0', (int) '3');
        Regex oct       = new CharRangeRegex((int) '0', (int) '7');
        Regex any       = new CharRangeRegex(0, (int) Character.MAX_VALUE);
        Regex expected = new SequenceRegex(
                Arrays.asList(
                        hash, backslash,
                        new AlternativeRegex(
                                Arrays.asList(
                                        any,
                                        new SequenceRegex(
                                                Arrays.asList(n, e, w, l, i,
                                                              n, e)),
                                        new SequenceRegex(
                                                Arrays.asList(s, p, a, c, e)),
                                        new SequenceRegex(
                                                Arrays.asList(t, a, b)),
                                        new SequenceRegex(
                                                Arrays.asList(
                                                        octFirst,
                                                        new RepeatRegex(
                                                                oct, 2,
                                                                2)))))));
        assertEquals(regex, expected);
    }

    @Test
    public void testString() throws IOException, RegexError {
        RegexParser parser = new RegexParser(
                "\"([^\\\"]|\\\\([tn]|[0-3][0-7]{2}))*\"");
        Regex regex = parser.run();
        Regex quote = new CharRangeRegex((int) '"', (int) '"');
        List<Pair<Integer, Integer>> normalCharRanges =
                CharRanges.negate(
                        Arrays.asList(
                                new Pair<>((int) '\\', (int) '\\'),
                                new Pair<>((int) '"', (int) '"')));
        Regex normalChar = new AlternativeRegex(
                normalCharRanges.stream()
                                .map(rng -> new CharRangeRegex(rng.left,
                                                               rng.right))
                                .collect(Collectors.toList()));
        Regex t         = new CharRangeRegex((int) 't', (int) 't');
        Regex n         = new CharRangeRegex((int) 'n', (int) 'n');
        Regex octFirst  = new CharRangeRegex((int) '0', (int) '3');
        Regex oct       = new CharRangeRegex((int) '0', (int) '7');
        Regex backslash = new CharRangeRegex((int) '\\', (int) '\\');
        Regex specialChar = new SequenceRegex(
                Arrays.asList(
                        backslash,
                        new AlternativeRegex(
                                Arrays.asList(
                                        new AlternativeRegex(
                                                Arrays.asList(t, n)),
                                        new SequenceRegex(
                                                Arrays.asList(
                                                        octFirst,
                                                        new RepeatRegex(
                                                                oct, 2,
                                                                2)))))));
        Regex expected = new SequenceRegex(
                Arrays.asList(
                        quote,
                        new RepeatRegex(
                                new AlternativeRegex(
                                        Arrays.asList(
                                                normalChar, specialChar)),
                                0, -1),
                        quote));
    }
}
